import typing

import fastapi
import pydantic

import cannula


class GraphQLPayload(pydantic.BaseModel):
    """Model representing a GraphQL request body."""

    query: str
    variables: typing.Optional[typing.Dict[str, typing.Any]] = None
    operationName: typing.Optional[str] = None


class ExecutionResponse(pydantic.BaseModel):
    data: typing.Optional[typing.Dict[str, typing.Any]] = None
    errors: typing.Optional[list] = None
    extensions: typing.Optional[typing.Dict[str, typing.Any]] = None


class GraphQLExec(typing.Protocol):
    async def __call__(
        self,
        context: typing.Any | None = None,
    ) -> ExecutionResponse: ...


class GraphQLDepends:
    """
    FastAPI Dependency to handle GraphQL requests.

    Example::

        from fastapi import FastAPI
        from cannula import CannulaAPI
        from cannula.contrib.asgi import ExecutionResponse, GraphQLDepends, GraphQLExec

        api = FastAPI()
        cannula_app = API()

        @api.post("/graph")
        async def _root(
            graph_call: typing.Annotated[
                GraphQLExec,
                Depends(GraphQLDepends(cannula_app)),
            ]
        ) -> ExecutionResponse:
            return await graph_call()

    This function allows you to pass context or the request in order to handle complex
    operations. For example you can setup `data_loaders` or a database session::

        ...
        async with async_session_maker() as session:
            context = {
                'session': session,
                'authorization': request.headers.get('authoriztion')
            }
            return await graph_call(context=context)
    """

    graph: cannula.CannulaAPI

    def __init__(self, graph: cannula.CannulaAPI) -> None:
        self.graph = graph

    async def __call__(
        self, request: fastapi.Request, payload: GraphQLPayload
    ) -> GraphQLExec:

        async def _call_graph(
            context: typing.Any | None = None,
        ) -> ExecutionResponse:
            results = await self.graph.call(
                document=payload.query,
                variables=payload.variables,
                operation_name=payload.operationName,
                context=context,
                request=request,
            )

            return ExecutionResponse(
                data=results.data,
                errors=cannula.format_errors(
                    results.errors,
                    logger=self.graph.logger,
                    level=self.graph.level,
                ),
                extensions=results.extensions,
            )

        return _call_graph



import typing
from dataclasses import fields

from sqlalchemy.orm import DeclarativeBase
from typing_extensions import Self


DBModel = typing.TypeVar("DBModel", bound=DeclarativeBase)


class DataclassInstance(typing.Protocol):
    __dataclass_fields__: typing.ClassVar[dict[str, typing.Any]]


class DBMixin(typing.Generic[DBModel]):
    """Extend dataclass to include a `from_db` factory function.

    This is useful if you are using something like sqlalchemy for your
    database access. With this decorator you can then initialize the
    GraphQL object classes with the model from your ORM. As long as
    the required fields are present. All extra fields from the model
    will be ignored.

    Example::

        from cannula.contrib.orm import DBMixin
        from models import Widget as DBWidget
        from ._generated import WidgetTypeBase

        class Widget(WidgetTypeBase, DBMixin[DBWidget]):
            pass

        async def get_widgets(info) -> list[Widget]
            db_objects = await select_from_database(DBWidget)
            return [Widget.from_db(obj) for obj in db_objects]

    The resulting init function will also continue to accept kwargs
    and can override a field that is in the database if needed::

        widget = Widget.from_db(db_model, name="something else")
        assert widget.name == "something else"
    """

    _db_model: DBModel

    @classmethod
    def from_db(cls, db_obj: DBModel, **kwargs) -> Self:
        model_kwargs = db_obj.__dict__
        model_kwargs.update(kwargs)
        expected_fields = {
            field.name for field in fields(typing.cast(DataclassInstance, cls))
        }
        cleaned_kwargs = {
            key: value for key, value in model_kwargs.items() if key in expected_fields
        }
        obj = cls(**cleaned_kwargs)
        obj._db_model = db_obj
        return obj
